package net.dries007.tfc.world.classic.chunkdata;

import net.dries007.tfc.objects.Ore;
import net.dries007.tfc.objects.blocks.BlockRockVariant;
import net.dries007.tfc.util.OreSpawnData;
import net.dries007.tfc.world.classic.DataLayer;
import net.minecraft.nbt.*;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.Constants;

import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import static net.dries007.tfc.world.classic.WorldTypeTFC.ROCKLAYER2;
import static net.dries007.tfc.world.classic.WorldTypeTFC.ROCKLAYER3;

@SuppressWarnings("WeakerAccess")
public final class ChunkDataTFC
{
    public static final int FISH_POP_MAX = 60;

    private static final ChunkDataTFC EMPTY = new ChunkDataTFC();

    private boolean initialized = false;
    private final DataLayer[] rockLayer1 = new DataLayer[256];
    private final DataLayer[] rockLayer2 = new DataLayer[256];
    private final DataLayer[] rockLayer3 = new DataLayer[256];
    private final DataLayer[] evtLayer = new DataLayer[256];
    private final DataLayer[] rainfallLayer = new DataLayer[256];
    private final DataLayer[] drainageLayer = new DataLayer[256];
    private final DataLayer[] stabilityLayer = new DataLayer[256];
    private final int[] seaLevelOffset = new int[256];

    private final List<ChunkDataOreSpawned> oresSpawned = new ArrayList<>();
    private final List<ChunkDataOreSpawned> oresSpawnedView = Collections.unmodifiableList(oresSpawned);

    private int fishPopulation = FISH_POP_MAX; // todo: Set this based on biome? temp? rng?

    static
    {
        Arrays.fill(EMPTY.rockLayer1, DataLayer.ERROR);
        Arrays.fill(EMPTY.rockLayer2, DataLayer.ERROR);
        Arrays.fill(EMPTY.rockLayer3, DataLayer.ERROR);
        Arrays.fill(EMPTY.evtLayer, DataLayer.ERROR);
        Arrays.fill(EMPTY.rainfallLayer, DataLayer.ERROR);
        Arrays.fill(EMPTY.drainageLayer, DataLayer.ERROR);
        Arrays.fill(EMPTY.stabilityLayer, DataLayer.ERROR);
        Arrays.fill(EMPTY.seaLevelOffset, -1);
    }

    public static ChunkDataTFC get(World world, BlockPos pos)
    {
        ChunkDataTFC data = world.func_175726_f(pos).getCapability(ChunkDataProvider.CHUNK_DATA_CAPABILITY, null);
        return data == null ? EMPTY : data;
    }

    public static BlockRockVariant getRock1(World world, BlockPos pos) { return get(world, pos).getRockLayer1(pos.func_177958_n() & 15, pos.func_177952_p() & 15).block; }
    public static BlockRockVariant getRock2(World world, BlockPos pos) { return get(world, pos).getRockLayer2(pos.func_177958_n() & 15, pos.func_177952_p() & 15).block; }
    public static BlockRockVariant getRock3(World world, BlockPos pos) { return get(world, pos).getRockLayer3(pos.func_177958_n() & 15, pos.func_177952_p() & 15).block; }
    public static float getEvt(World world, BlockPos pos) { return get(world, pos).getEvtLayer(pos.func_177958_n() & 15, pos.func_177952_p() & 15).valueFloat; }
    public static float getRainfall(World world, BlockPos pos) { return get(world, pos).getRainfallLayer(pos.func_177958_n() & 15, pos.func_177952_p() & 15).valueFloat; }
    public static boolean isStable(World world, BlockPos pos) { return get(world, pos).getStabilityLayer(pos.func_177958_n() & 15, pos.func_177952_p() & 15).valueInt == 0; }
    public static int getDrainage(World world, BlockPos pos) { return get(world, pos).getDrainageLayer(pos.func_177958_n() & 15, pos.func_177952_p() & 15).valueInt; }
    public static int getSeaLevelOffset(World world, BlockPos pos) { return get(world, pos).getSeaLevelOffset(pos.func_177958_n() & 15, pos.func_177952_p() & 15); }
    public static int getFishPopulation(World world, BlockPos pos) { return get(world, pos).getFishPopulation(); }
    public static BlockRockVariant getRockHeight(World world, BlockPos pos) { return get(world, pos).getRockLayerHeight(pos.func_177958_n() & 15, pos.func_177956_o(), pos.func_177952_p() & 15).block; }

    /**
     * INTERNAL USE ONLY.
     * No need to mark as dirty, since this will only ever be called on worldgen, before the first chunk save.
     */
    public void setGenerationData(DataLayer[] rockLayer1, DataLayer[] rockLayer2, DataLayer[] rockLayer3, DataLayer[] evtLayer, DataLayer[] rainfallLayer, DataLayer[] stabilityLayer, DataLayer[] drainageLayer, int[] seaLevelOffset)
    {
        this.initialized = true;
        System.arraycopy(rockLayer1, 0, this.rockLayer1, 0, 256);
        System.arraycopy(rockLayer2, 0, this.rockLayer2, 0, 256);
        System.arraycopy(rockLayer3, 0, this.rockLayer3, 0, 256);
        System.arraycopy(evtLayer, 0, this.evtLayer, 0, 256);
        System.arraycopy(rainfallLayer, 0, this.rainfallLayer, 0, 256);
        System.arraycopy(stabilityLayer, 0, this.stabilityLayer, 0, 256);
        System.arraycopy(drainageLayer, 0, this.drainageLayer, 0, 256);
        System.arraycopy(seaLevelOffset, 0, this.seaLevelOffset, 0, 256);
    }

    /**
     * INTERNAL USE ONLY.
     */
    public void addSpawnedOre(Ore ore, OreSpawnData.SpawnSize size, Ore.Grade grade, BlockPos pos, int count)
    {
        oresSpawned.add(new ChunkDataOreSpawned(ore, size, grade, pos, count));
    }

    public boolean isInitialized()
    {
        return initialized;
    }

    public BlockRockVariant getRock1(int x, int z) { return getRockLayer1(x, z).block; }
    public BlockRockVariant getRock2(int x, int z) { return getRockLayer2(x, z).block; }
    public BlockRockVariant getRock3(int x, int z) { return getRockLayer3(x, z).block; }
    public float getEvt(int x, int z) { return getEvtLayer(x, z).valueFloat; }
    public float getRainfall(int x, int z) { return getRainfallLayer(x, z).valueFloat; }
    public boolean isStable(int x, int z) { return getStabilityLayer(x, z).valueInt == 0; }
    public int getDrainage(int x, int z) { return getDrainageLayer(x, z).valueInt; }
    public BlockRockVariant getRockHeight(int x, int y, int z) { return getRockLayerHeight(x & 15, y, z & 15).block; }

    public int getSeaLevelOffset(int x, int z) { return seaLevelOffset[z << 4 | x]; }

    public int getFishPopulation() { return fishPopulation; }
    public List<ChunkDataOreSpawned> getOresSpawned() { return oresSpawnedView; }

    // Directly accessing the DataLayer is discouraged (except for getting the name). It's easy to use the wrong value.
    public DataLayer getRockLayer1(int x, int z) {      return rockLayer1   [z << 4 | x]; }
    public DataLayer getRockLayer2(int x, int z) {      return rockLayer2   [z << 4 | x]; }
    public DataLayer getRockLayer3(int x, int z) {      return rockLayer3   [z << 4 | x]; }
    public DataLayer getEvtLayer(int x, int z) {        return evtLayer     [z << 4 | x]; }
    public DataLayer getRainfallLayer(int x, int z) {   return rainfallLayer[z << 4 | x]; }
    public DataLayer getStabilityLayer(int x, int z) { return stabilityLayer[z << 4 | x]; }
    public DataLayer getDrainageLayer(int x, int z) {   return drainageLayer[z << 4 | x]; }

    public DataLayer getRockLayerHeight(int x, int y, int z)
    {
        int offset = getSeaLevelOffset(x, z);
        if (y <= ROCKLAYER3 + offset) return getRockLayer3(x, z);
        if (y <= ROCKLAYER2 + offset) return getRockLayer2(x, z);
        return getRockLayer1(x, z);
    }

    public static final class ChunkDataStorage implements Capability.IStorage<ChunkDataTFC>
    {
        public static NBTTagByteArray write(DataLayer[] layers)
        {
            return new NBTTagByteArray(Arrays.stream(layers).map(x -> (byte)x.layerID).collect(Collectors.toList()));
        }

        public static void read(DataLayer[] layers, byte[] bytes)
        {
            for (int i = bytes.length - 1; i >= 0; i--)
            {
                layers[i] = DataLayer.get(bytes[i]);
            }
        }

        @Nullable
        @Override
        public NBTBase writeNBT(Capability<ChunkDataTFC> capability, ChunkDataTFC instance, EnumFacing side)
        {
            if (instance == null) return null;
            NBTTagCompound root = new NBTTagCompound();
            root.func_74782_a("rockLayer1", write(instance.rockLayer1));
            root.func_74782_a("rockLayer2", write(instance.rockLayer2));
            root.func_74782_a("rockLayer3", write(instance.rockLayer3));
            root.func_74782_a("evtLayer", write(instance.evtLayer));
            root.func_74782_a("rainfallLayer", write(instance.rainfallLayer));
            root.func_74782_a("stabilityLayer", write(instance.stabilityLayer));
            root.func_74782_a("drainageLayer", write(instance.drainageLayer));

            root.func_74782_a("seaLevelOffset", new NBTTagIntArray(instance.seaLevelOffset));
            root.func_74768_a("fishPopulation", instance.fishPopulation);

            NBTTagList chunkDataOreSpawnedNBT = new NBTTagList();
            instance.oresSpawned.stream().map(ChunkDataOreSpawned::serialize).forEach(chunkDataOreSpawnedNBT::func_74742_a);
            root.func_74782_a("oresSpawned", chunkDataOreSpawnedNBT);

            return root;
        }

        @Override
        public void readNBT(Capability<ChunkDataTFC> capability, ChunkDataTFC instance, EnumFacing side, NBTBase nbt)
        {
            NBTTagCompound root = ((NBTTagCompound) nbt);
            read(instance.rockLayer1, root.func_74770_j("rockLayer1"));
            read(instance.rockLayer2, root.func_74770_j("rockLayer2"));
            read(instance.rockLayer3, root.func_74770_j("rockLayer3"));
            read(instance.evtLayer, root.func_74770_j("evtLayer"));
            read(instance.rainfallLayer, root.func_74770_j("rainfallLayer"));
            read(instance.stabilityLayer, root.func_74770_j("stabilityLayer"));
            read(instance.drainageLayer, root.func_74770_j("drainageLayer"));

            System.arraycopy(root.func_74759_k("seaLevelOffset"), 0, instance.seaLevelOffset, 0, 256);
            instance.fishPopulation = root.func_74762_e("fishPopulation");

            instance.oresSpawned.clear();
            root.func_150295_c("oresSpawned", Constants.NBT.TAG_COMPOUND).forEach(x -> instance.oresSpawned.add(new ChunkDataOreSpawned(((NBTTagCompound) x))));

            instance.initialized = true;
        }
    }
}
